/*

	Filename:   MemoText.cfm
   Version:    1.30
   Date:       2001 03 28

   For:        VdBase 7.x / dB2K / Win9x / Win2000

   Author:     Ronnie MacGregor
   Contact:    ronnie@high-lugtonridge.co.uk
	
   
   Revisions:
   **********
      2001 06 09  Automatic creation of MemoText table option added to 
                  mtSetup().
      2001 08 13  Bug fix to deal with multiple rowsets containing MemoText.
                  There is now an individual MemoText field array attached 
                  to each participating rowset, to ensure that saving and 
                  deleting of MemoText is only attempted for fields in
                  the appropriate rowset.
                        
         
   Description:
   ************
      A custom Form for memo text which avoids using a blob file
      (.dbt) for storage, by storing the memo text in table rows.


   Methods:
   ********
      mtSetup(oRowsetRef,cTableName)

         Should be called from the onOpen event of the query for the rowset
         containing the link field for the MemoText.

         oRowsetRef

            An object reference to the query's rowset.

         cTableName
            
            A string containing the name of the table to store MemoText in.            


      mtBuildMemo()

         This method can be called by the beforeGetValue event of a calculated
         field in order to populate the field with MemoText. The calculated
         field requires two custom properties, mtLinkField and mtTableName,
         which the method uses to retrieve the MemoText.


      Generate_ID

         A function which returns a unique value, which is used here as a link
         for MemoText.


      mtOnChange

         This function is called by the calculated field to ensure that the rowset
         is aware that there has been data change. 


      mtCanSave and mtCanDelete

         Functions which are called by the rowset containing the link field to the
         MemoText. These functions look after the saving and deleting of MemoText.        
                                              

   Dependancies:
   *************
      Requires a table with the following structure to store the Memo Text.

         Memo_ID         Character   10  .0   None
         Memo_Sequence   Character    4  .0   None
         Memo_Text       Character  100  .0   None

      Requires that the rowset events for the table containing the link field to
      the memo text are set up to fire the cfm's corresponding methods.
      Along these lines:

         canSave = class::mtCanSave
         canDelete = class::mtCanDelete

      Calculated fields must be created in the query's onOpen event, to add
      MemoText fields the rowset containing the link field. All of the following
      custom field properties and events must be set up.

         f = new Field()
         f.fieldName = "MemoText1"
         f.mtLinkField = "MemoText1_ID" 
         f.mtTableName = "mtOrders.mtf" 
         f.mtLinkFieldPointer = this.rowset.fields["MemoText1_ID"]
         f.beforeGetValue = class::mtBuildMemo
         f.onChange = class::mtOnChange
         this.rowset.fields.add( f )
                
      The standard editor used must give the calculated field a reference to
      itself in it's onOpen event by creating the following custom field
      property, along these lines:

.         function ED_MT1_onOpen
            // Give the field a reference to this control
            form.mtordersquery.rowset.fields["memotext1"];
            .VisualControlRef = this
            return
            

   Conditions of use:
   ******************
      This cfm is offered as "freeware" to dBASE and dB2K developers,
      but must be credited to the author.

      Since the use of this control is in data storage, it is the duty of the
      user to determine it's suitability for use. In no circumstance will the
      author accept any responsibility for any data loss, or any other problem
      which might arise from the use of this control.

      Feedback:

      If you find this control useful, the author would appreciate an e-mail
      to that effect.

      If you find any bugs or problems, the author would appreciate an e-mail
      with concise details. 

         
*/

** END HEADER -- do not remove this line
//
// Generated on 28/03/2001
//
class MemoTextBaseForm of FORM custom

   
   function mtSetup(oRowsetRef,cTableName)
      // Check for MemoText table and create if required
      if not file(cTableName)
         if msgbox("The MemoText table " + cTableName + " cannot be found.";
          + chr(13) + "Do you wish a new empty table to be created ?","Missing Table",52) = 6
            private cNewTable
            cNewTable = "create table " + '"' + cTableName + '"' + " (";
                 + "Memo_ID        char(10),";
                 + "Memo_Sequence  char(4),";
                 + "Memo_Text      char(100))"
            &cNewTable.
          else
             msgbox("You cannot continue error free without this table","Missing Table",16)
          endif
      endif
      //
      // Check first if array has already been created.
      try
         x = oRowsetRef.mtFieldArray.size
         x = null
      catch(Exception e)
         oRowsetRef.mtFieldArray = new array()
      endtry
      //
      // Search for MemoText fields 
      for i = 1 to oRowsetRef.fields.size
         try
            x = oRowsetRef.fields[i].mtLinkFieldPointer
            x = oRowsetRef.fields[i].mtLinkField
            x = oRowsetRef.fields[i].mtTableName
            x = null
            oRowsetRef.mtFieldArray.add(oRowsetRef.fields[i])
         catch(Exception e)
            // do nothing - not a MemoText field
         endtry
      next
   return


   function mtBuildMemo()
      // Used to populate a calculated field with Memotext.
      //
      // Calculated field object must have valid mtLinkField and mtTableName
      // properties attached (as strings).
      //
      local sConcant, tpp
      sConcant = ""
      tpp = this.parent.parent
      //
      if tpp.fields[this.mtLinkField].value = null
         return sConcant
      endif
      //
      if not tpp.endofset
         // Set up Query
         sQueryString = "'select * from " ;
            + '"' + this.mtTableName + '"';
            + " where Memo_ID = " ;
            + '"' + tpp.fields[this.mtLinkField].value + '"' ;
            + "'"
         //
         this.q = new Query()
         this.q.parent = this
         with (this.q)
            sql = &sQueryString.
            active = true
         endwith
         //
         // Build memo contents
         this.q.rowset.first()
         do while not this.q.rowset.endofset
            sConcant = sConcant + this.q.Rowset.Fields["Memo_Text"].value
            this.q.Rowset.next()
         enddo
         // Close query
         this.q.active = false
         this.q = null
         //
      endif
      //
      return sConcant
            
            
   function mtOnChange()
      //
      // Called by field to ensure rowset save events fire.
      this.parent.parent.modified = true
      return


   function mtCanSave()
      local sConcant, tpp
      tpp = this.parent.parent
      //
      for i = 1 to this.mtFieldArray.size
         //
         // Deal with no pre-existing memo entry for this field
         if this.mtFieldArray[i].mtLinkFieldPointer.value = null
            this.mtFieldArray[i].mtLinkFieldPointer.value = ""
         endif
         if len(trim(this.mtFieldArray[i].mtLinkFieldPointer.value)) = 0 ;
            and len(trim(this.mtFieldArray[i].VisualControlRef.value)) > 0
            this.mtFieldArray[i].mtLinkFieldPointer.value = tpp.Generate_ID()
         endif
         //
         // Set up Query
         sQueryString = "'select * from " ;
            + '"' + this.mtFieldArray[i].mtTableName + '"';
            + " where Memo_ID = " ;
            + '"' + this.mtFieldArray[i].mtLinkFieldPointer.value + '"' ;
            + "'"
         //
         this.q = new Query()
         this.q.parent = this
         with (this.q)
            sql = &sQueryString.
            active = true
         endwith
         //
         // Get text from control to save
         try
            sConcant = trim(this.mtFieldArray[i].VisualControlRef.value)
         catch(Exception e)
            msgbox("Missing VisualControlRef property.","Setup error",16)
            return false
         endtry
         //
         // Deal with deleted memo contents
         if len(trim(sConcant)) = 0
            this.mtFieldArray[i].mtLinkFieldPointer.value = ""
         endif
         //
         if not this.q.rowset.state = 3 // append
            this.q.rowset.first()
         endif
         //
         local nSeq, sMemoChunk
         nSeq = 1
         //
         // Start saving
         do while len(sConcant) > 0
            if this.q.rowset.endofset
               this.q.rowset.beginAppend()
            endif
            this.q.rowset.Fields["Memo_ID"].value = ;
               this.mtFieldArray[i].mtLinkFieldPointer.value
            this.q.rowset.Fields["Memo_Sequence"].value = str(nSeq,4,0,"0")
            sMemoChunk = substr(sConcant,1,100)
            this.q.rowset.Fields["Memo_Text"].value = sMemoChunk
            sConcant = substr(sConcant,101,len(sConcant))
            if len(sConcant) < 100
               sConcant = trim(sConcant)
            endif
            this.q.rowset.save()
            this.q.rowset.next()
            nSeq = nSeq + 1
         enddo
         //
         // Clean up unused rows (for shorter memo)
         do while not this.q.rowset.endofset
            this.q.rowset.delete() // this moves us to the next row
         enddo
         //
         // Close query
         this.q.active = false
         this.q = null
         //
      next
      //
      return true


   function mtCanDelete()
      local tpp
      tpp = this.parent.parent
      //
      for i = 1 to this.mtFieldArray.size
         // Set up Query
         sQueryString = "'select * from " ;
            + '"' + this.mtFieldArray[i].mtTableName + '"';
            + " where Memo_ID = " ;
            + '"' + this.mtFieldArray[i].mtLinkFieldPointer.value + '"' ;
            + "'"
         //
         this.q = new Query()
         this.q.parent = this
         with (this.q)
            sql = &sQueryString.
            active = true
         endwith
         //
         // Delete all matching rows
         this.q.rowset.first()
         do while not this.q.rowset.endofset
            this.q.rowset.delete() // this moves us to the next row
         enddo
         //
         // Close query
         this.q.active = false
         this.q = null
         //
      next
      //
      // Try to pack table
      try
         _app.databases[1].packTable(this.mtFieldArray[1].mtTableName)
      catch ( Exception e )
         // Do nothing - can't pack table without exclusive use.
         // Open project file prevents this in development environment.
      endTry
      //
      return true
            

   function Generate_ID()
      return itoh(val(year(date()) ;
         + str(month(date()),2,0,"0") ;
         + str(day(date()),2,0,"0") ;
         + str(int((seconds() * 100)),7,0,"0")),8) ;
         + str((random()*100),2,0,"0")

                    
endclass
